{*******************************************************}
{                                                       }
{                 "Virtual Town" project                }
{                                                       }
{               Copyright (c) 2006 PAS             }
{                                                       }
{Authors: Alex Ponomarev <AlxPonom@mail.ru>             }
{                                                       }
{  This program is free software under the GPL (>= v2)  }
{ Read the file COPYING coming with project for details }
{*******************************************************}

unit unitHugeBitmapDivider;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ExtCtrls, StdCtrls, Buttons, ExtDlgs;


type
  TfmHugeBitmapDivider = class(TForm)
    edBitmapFileName: TEdit;
    sbSelectFileName: TSpeedButton;
    Label1: TLabel;
    Bevel1: TBevel;
    sbStart: TSpeedButton;
    ProgressBar: TProgressBar;
    OpenDialog: TOpenDialog;
    edOutputFolder: TEdit;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    lbEmptySegmentColor: TLabel;
    lbEmptySegmentBorderColor: TLabel;
    sbEmptySegmentColor: TSpeedButton;
    sbEmptySegmentBorderColor: TSpeedButton;
    ColorDialog: TColorDialog;
    procedure sbStartClick(Sender: TObject);
    procedure sbSelectFileNameClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure sbEmptySegmentColorClick(Sender: TObject);
    procedure sbEmptySegmentBorderColorClick(Sender: TObject);
  private
    { Private declarations }
    OutputFolder: string;
    
    procedure DivideHugeBitmapIntoTiles(const FileName: TFileName; const SegmentSize: integer; const EmptySegmentColor,EmptySegmentBorderColor: TColor; const OutputFolder: string);
  public
    { Public declarations }
  end;

var
  fmHugeBitmapDivider: TfmHugeBitmapDivider;

implementation
Uses
  Jpeg;

{$R *.dfm}


procedure TfmHugeBitmapDivider.DivideHugeBitmapIntoTiles(const FileName: TFileName; const SegmentSize: integer; const EmptySegmentColor,EmptySegmentBorderColor: TColor; const OutputFolder: string);
const
  BackGroundColor = clGray;

  function GetSize(const IniSize: integer; out Level: integer): integer;
  begin
  Result:=SegmentSize;
  Level:=0;
  while (Result < IniSize) do begin
    Result:=(Result SHL 1);
    Inc(Level);
    end;
  end;

var
  BMPFile: TFileStream;
  BitmapHeader: TBitmapFileHeader;
  BitmapInfo: TBitmapInfo;
  FilePosition: integer;
  Bitmap: TBitmap;
  Yoffset: integer;
  Jpeg: TJpegImage;
  X,Y: integer;
  Xindex,Yindex: integer;
  SizeX: integer;
  I,J: integer;
  Spacing: Integer;
  D,Dr: integer;
  L,Lr: integer;
  SegmentsLength: integer;
  SegmentFileName: string;
begin
Bitmap:=TBitmap.Create;
try
Bitmap.HandleType:=bmDIB;
Bitmap.PixelFormat:=pf24Bit;
Bitmap.Width:=SegmentSize;
Bitmap.Height:=SegmentSize;
Jpeg:=TJpegImage.Create;
try
BMPFile:=TFileStream.Create(FileName,fmOpenRead);
try
BMPFile.Read(BitmapHeader,SizeOf(BitmapHeader));
BMPFile.Read(BitmapInfo,SizeOf(BitmapInfo));
if (BitmapInfo.bmiHeader.biBitCount <> 24) then Raise Exception.Create('Only 24 bit bitmap is supported'); //. =>
Spacing:=(BitmapInfo.bmiHeader.biWidth MOD 4);
FilePosition:=(BMPFile.Position+2);
//.
Dr:=GetSize(BitmapInfo.bmiHeader.biWidth,Lr);
D:=GetSize(BitmapInfo.bmiHeader.biHeight,L);
if (D > Dr)
 then begin
  Dr:=D;
  Lr:=L;
  end;
SegmentsLength:=(1 SHL Lr);
//.
ForceDirectories(OutputFolder);
Bitmap.Canvas.Brush.Color:=EmptySegmentColor;
Bitmap.Canvas.Pen.Color:=EmptySegmentBorderColor;
ProgressBar.Max:=SegmentsLength;
Y:=BitmapInfo.bmiHeader.biHeight-SegmentSize;
Yindex:=0;
while ((Y+SegmentSize) > 0) do begin
  if (Y >= 0)
   then Yoffset:=0
   else begin
    Yoffset:=-Y;
    Y:=0;
    end;
  Xindex:=0;
  X:=0;
  while (X < BitmapInfo.bmiHeader.biWidth) do begin
    SizeX:=(BitmapInfo.bmiHeader.biWidth-X); if (SizeX > SegmentSize) then SizeX:=SegmentSize;
    //.
    Bitmap.Canvas.Rectangle(0,0,Bitmap.Width,Bitmap.Height);
    //.
    BMPFile.Position:=FilePosition+(((BitmapInfo.bmiHeader.biWidth*3)+Spacing)*Y)+(X*3);
    for I:=SegmentSize-1 downto Yoffset do begin
      BMPFile.Read(Bitmap.ScanLine[I-Yoffset]^,SizeX*3);
      BMPFile.Position := BMPFile.Position+((BitmapInfo.bmiHeader.biWidth-SizeX)*3)+Spacing;
      end;
    Jpeg.Assign(Bitmap);
    SegmentFileName:=OutputFolder+'\'+'Segment_'+IntToStr(Lr)+'_'+IntToStr(Xindex)+'_'+IntToStr(Yindex)+'.jpg';
    Jpeg.SaveToFile(SegmentFileName);
    //.
    Inc(X,SegmentSize);
    Inc(Xindex);
    end;
  //.
  Bitmap.Canvas.Rectangle(0,0,Bitmap.Width,Bitmap.Height);
  while (Xindex < SegmentsLength) do begin
    Jpeg.Assign(Bitmap);
    SegmentFileName:=OutputFolder+'\'+'Segment_'+IntToStr(Lr)+'_'+IntToStr(Xindex)+'_'+IntToStr(Yindex)+'.jpg';
    Jpeg.SaveToFile(SegmentFileName);
    //.
    Inc(Xindex);
    end;
  //.
  ProgressBar.Position:=Yindex;
  //.
  Dec(Y,SegmentSize);
  Inc(Yindex);
  end;
while(YIndex < SegmentsLength) do begin
  Xindex:=0;
  while (Xindex < SegmentsLength) do begin
    Jpeg.Assign(Bitmap);
    SegmentFileName:=OutputFolder+'\'+'Segment_'+IntToStr(Lr)+'_'+IntToStr(Xindex)+'_'+IntToStr(Yindex)+'.jpg';
    Jpeg.SaveToFile(SegmentFileName);
    //.
    Inc(Xindex);
    end;
  //.
  ProgressBar.Position:=Yindex;
  //.
  Inc(Yindex);
  end;
ProgressBar.Position:=ProgressBar.Max;
finally
BMPFile.Destroy;
end;
finally
Jpeg.Destroy;
end;
finally
Bitmap.Destroy;
end;
ShowMessage('Bitmap was divided successfully'#$0D#$0A'"Size power of 2" = '+IntToStr(Lr)+' (remember this number for a segmented image generation).');
end;


procedure TfmHugeBitmapDivider.sbStartClick(Sender: TObject);
begin
DivideHugeBitmapIntoTiles(edBitmapFileName.Text,256,lbEmptySegmentColor.Color,lbEmptySegmentBorderColor.Color,OutputFolder);
end;

procedure TfmHugeBitmapDivider.sbSelectFileNameClick(Sender: TObject);
begin
if (OpenDialog.Execute()) then edBitmapFileName.Text:=OpenDialog.FileName;
end;

procedure TfmHugeBitmapDivider.FormCreate(Sender: TObject);
var
  I: integer;
begin
I:=1;
while (DirectoryExists(ExtractFilePath(ParamStr(0))+'Output'+'\'+'Tiles'+IntToStr(I))) do Inc(I);
OutputFolder:=ExtractFilePath(ParamStr(0))+'Output'+'\'+'Tiles'+IntToStr(I);
edOutputFolder.Text:=OutputFolder;
end;

procedure TfmHugeBitmapDivider.sbEmptySegmentColorClick(Sender: TObject);
begin
if (ColorDialog.Execute()) then lbEmptySegmentColor.Color:=ColorDialog.Color;
end;

procedure TfmHugeBitmapDivider.sbEmptySegmentBorderColorClick(
  Sender: TObject);
begin
if (ColorDialog.Execute()) then lbEmptySegmentBorderColor.Color:=ColorDialog.Color;
end;

end.
